@@ -24,78 +24,49 @@
 
 namespace cppbugs {
 
-  ////////////////////////////////////////////////////////////////////////////////////
-  // ICSIlog Copyright taken from ICSIlog source                                    //
-  // Copyright (C) 2007 International Computer Science Institute                    //
-  // 1947 Center Street, Suite 600                                                  //
-  // Berkeley, CA 94704                                                             //
-  //                                                                                //
-  // Contact information:                                                           //
-  //    Oriol Vinyals	vinyals@icsi.berkeley.edu                                   //
-  //    Gerald Friedland 	fractor@icsi.berkeley.edu                           //
-  //                                                                                //
-  // This program is free software; you can redistribute it and/or modify           //
-  // it under the terms of the GNU General Public License as published by           //
-  // the Free Software Foundation; either version 2 of the License, or              //
-  // (at your option) any later version.                                            //
-  //                                                                                //
-  // This program is distributed in the hope that it will be useful,                //
-  // but WITHOUT ANY WARRANTY; without even the implied warranty of                 //
-  // MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  //
-  // GNU General Public License for more details.                                   //
-  //                                                                                //
-  // You should have received a copy of the GNU General Public License along        //
-  // with this program; if not, write to the Free Software Foundation, Inc.,        //
-  // 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.                    //
-  //                                                                                //
-  // Authors                                                                        //
-  // -------                                                                        //
-  //                                                                                //
-  // Oriol Vinyals		vinyals@icsi.berkeley.edu                           //
-  // Gerald Friedland 	fractor@icsi.berkeley.edu                                   //
-  //                                                                                //
-  // Acknowledgements                                                               //
-  // ----------------                                                               //
-  //                                                                                //
-  // Thanks to Harrison Ainsworth (hxa7241@gmail.com) for his idea that             //
-  // doubled the accuracy.                                                          //
-  ////////////////////////////////////////////////////////////////////////////////////
+  arma_hot arma_pure arma_inline
+  float log_approx(float val) {
+    union { float f; int i; } valu;
+    valu.f = val;
+    float exp = valu.i >> 23;
 
+    float addcst = val > 0 ? -89.76031805111f : -INFINITY;
 
-  /* ICSIlog V 2.0 */
-  const std::vector<float> fill_icsi_log_table2(const unsigned int precision) {
-    std::vector<float> pTable(1<<precision);
+    valu.i = (valu.i & 0x7FFFFF) | 0x3F800000;
 
-    /* step along table elements and x-axis positions
-       (start with extra half increment, so the steps intersect at their midpoints.) */
-    float oneToTwo = 1.0f + (1.0f / (float)( 1 <<(precision + 1) ));
-    for(int i = 0;  i < (1 << precision);  ++i ) {
-      // make y-axis value for table element
-      pTable[i] = logf(oneToTwo) / 0.69314718055995f;
-      oneToTwo += 1.0f / (float)( 1 << precision );
-    }
-    return pTable;
-  }
+    float x = valu.f;
 
-  /* ICSIlog v2.0 */
-  inline double icsi_log(const double vald) {
-    const float val = static_cast<float>(vald);
-    const unsigned int precision(10);
-    static std::vector<float> pTable = fill_icsi_log_table2(precision);
-
-    /* get access to float bits */
-    const int* const pVal = (const int*)(&val);
-
-    /* extract exponent and mantissa (quantized) */
-    const int exp = ((*pVal >> 23) & 255) - 127;
-    const int man = (*pVal & 0x7FFFFF) >> (23 - precision);
-
-    /* exponent plus lookup refinement */
-    return static_cast<double>(((float)(exp) + pTable[man]) * 0.69314718055995f);
+    return x*(2.79224171f+x*(-1.44246914f+x*(0.43585738f+x*-0.05486225f)))
+      + (addcst + 0.69314718055995f*exp);
   }
 
   inline double log_approx(const double x) {
-    return x <= 0 ? -std::numeric_limits<double>::infinity() : icsi_log(x);
+    return log_approx((float)x);
+  }
+
+  inline float log_approx(const int x) {
+    return log_approx((float)x);
+  }
+
+  // We do not inline these constants, because that makes GCC
+  // not able to recognize max/min pattern, and then the code is
+  // not vectorized.
+  float exp_cst1 = 2139095040.f;
+  float exp_cst2 = 0.f;
+
+  arma_hot arma_pure arma_inline
+  float exp_approx(float val) {
+    union { int i; float f; } xu;
+    float val2 = 12102203.1615614f*val+1065353216.f;
+    float val3 = val2 < exp_cst1 ? val2 : exp_cst1;
+    float val4 = val3 > exp_cst2 ? val3 : exp_cst2;
+    int val4i = (int) val4;
+    xu.i = val4i & 0x7F800000;
+    union { int i; float f; } xu2;
+    xu2.i = (val4i & 0x7FFFFF) | 0x3F800000;
+    float b = xu2.f;
+    float ret = xu.f * (0.51079604f+b*(0.30980503f+b*(0.16876894f+b*(-0.00303925f+b*0.01367652f))));
+    return ret;
   }
 }
 
@@ -125,6 +96,31 @@ namespace arma {
   }
 }
 
+namespace arma {
+  // exp_approx
+  class eop_exp_approx: public eop_core<eop_exp_approx> {};
+
+  template<> template<typename eT> arma_hot arma_pure arma_inline eT
+  eop_core<eop_exp_approx>::process(const eT val, const eT  ) {
+    return cppbugs::exp_approx(val);
+  }
+
+  // Base
+  template<typename T1>
+  arma_inline
+  const eOp<T1, eop_exp_approx> exp_approx(const Base<typename T1::elem_type,T1>& A) {
+    arma_extra_debug_sigprint();
+    return eOp<T1, eop_exp_approx>(A.get_ref());
+  }
+
+  // BaseCube
+  template<typename T1>
+  arma_inline
+  const eOpCube<T1, eop_exp_approx> exp_approx(const BaseCube<typename T1::elem_type,T1>& A) {
+    arma_extra_debug_sigprint();
+    return eOpCube<T1, eop_exp_approx>(A.get_ref());
+  }
+}
 
 namespace arma {
   // lgamma
@@ -155,7 +151,6 @@ namespace arma {
 
 namespace arma {
   // factln
-
   double factln(const int i) {
     static std::vector<double> factln_table;
 
@@ -253,13 +248,18 @@ namespace arma {
 
   double schur(const int x, const double y) { return x * y; }
   double schur(const double x, const int y) { return x * y; }
-  double schur(const double& x, const double& y) { return x * y; }
-  double schur(const int& x, const int& y) { return x * y; }
+  double schur(const double x, const double y) { return x * y; }
+  double schur(const float x, const double y) { return x * y; }
+  double schur(const double x, const float y) { return x * y; }
+  float schur(const int x, const float y) { return x * y; }
+  float schur(const float x, const int y) { return x * y; }
+  float schur(const float x, const float y) { return x * y; }
+  int schur(const int x, const int y) { return x * y; }
 
 }
 
 namespace arma {
-  bool any(const bool x) {
+  bool all(const bool x) {
     return x;
   }
 }
@@ -271,6 +271,10 @@ namespace cppbugs {
     return 1;
   }
 
+  double dim_size(const float) {
+    return 1;
+  }
+
   double dim_size(const int) {
     return 1;
   }
@@ -288,6 +292,10 @@ namespace cppbugs {
     return x*x;
   }
 
+  static inline float square(float x) {
+    return x*x;
+  }
+
   static inline int square(int x) {
     return x*x;
   }
@@ -304,44 +312,44 @@ namespace cppbugs {
 
   template<typename T, typename U, typename V>
   double normal_logp(const T& x, const U& mu, const V& tau) {
-    return arma::accu(0.5*log_approx(0.5*tau/arma::math::pi()) - 0.5 * arma::schur(tau, square(x - mu)));
+    return arma::accu(0.5f*log_approx(0.5f*tau/arma::math::pi()) - 0.5f * arma::schur(tau, square(x - mu)));
   }
 
   template<typename T, typename U, typename V>
   double uniform_logp(const T& x, const U& lower, const V& upper) {
-    return arma::any((x < lower) + (x > upper)) ? -std::numeric_limits<double>::infinity() : -arma::accu(log_approx(upper - lower));
+    if(!arma::all(x > lower) || !arma::all(x < upper))
+      return -std::numeric_limits<double>::infinity();
+    return -arma::accu(log_approx(upper - lower));
   }
 
   template<typename T, typename U, typename V>
   double gamma_logp(const T& x, const U& alpha, const V& beta) {
-    return arma::any(x < 0) ?
-      -std::numeric_limits<double>::infinity() :
-      arma::accu(arma::schur((alpha - 1.0),log_approx(x)) - arma::schur(beta,x) - lgamma(alpha) + arma::schur(alpha,log_approx(beta)));
+    if(!arma::all(x > 0))
+      return -std::numeric_limits<double>::infinity();
+    return
+      arma::accu(arma::schur((alpha - 1.0f),log_approx(x)) - arma::schur(beta,x) - lgamma(alpha) + arma::schur(alpha,log_approx(beta)));
   }
 
   template<typename T, typename U, typename V>
   double beta_logp(const T& x, const U& alpha, const V& beta) {
-    return arma::any((x <= 0 ) + (x >= 1)) ||
-      arma::any(alpha <= 0) || arma::any(beta <= 0) ?
-      -std::numeric_limits<double>::infinity() :
-      arma::accu(lgamma(alpha+beta) - lgamma(alpha) - lgamma(beta) + arma::schur(alpha - 1.0, log_approx(x)) + arma::schur(beta - 1.0, log_approx(1.0 - x)));
+    if(!arma::all(x > 0) || !arma::all(x < 1) ||
+       !arma::all(alpha > 0) || !arma::all(beta > 0))
+      return -std::numeric_limits<double>::infinity();
+    return arma::accu(lgamma(alpha+beta) - lgamma(alpha) - lgamma(beta) + arma::schur(alpha - 1.0f, log_approx(x)) + arma::schur(beta - 1.0f, log_approx(1.0f - x)));
   }
 
   template<typename T, typename U, typename V>
   double binom_logp(const T& x, const U& n, const V& p) {
-    if(arma::any((p <= 0) + (p >= 1)) || arma::any((x < 0) + (x > n))) {
+    if(!arma::all(x >= 0) || !arma::all(x <= n))
       return -std::numeric_limits<double>::infinity();
-    }
     return arma::accu(arma::schur(x,log_approx(p)) + arma::schur((n-x),log_approx(1-p)) + arma::factln(n) - arma::factln(x) - arma::factln(n-x));
   }
 
   template<typename T, typename U>
   double bernoulli_logp(const T& x, const U& p) {
-    if( arma::any((p <= 0) + (p >= 1)) || arma::any((x < 0) + (x > 1))) {
+    if(!arma::all(x >= 0) || !arma::all(x <= 1))
       return -std::numeric_limits<double>::infinity();
-    } else {
-      return arma::accu(arma::schur(x,log_approx(p)) + arma::schur((1-x), log_approx(1-p)));
-    }
+    return arma::accu(arma::schur(x,log_approx(p)) + arma::schur((1-x), log_approx(1-p)));
   }
 
   // sigma denotes cov matrix rather than precision matrix
