@@ -3,85 +3,40 @@
 #define BOOST_DISABLE_ASSERTS
 #define ARMA_DONT_USE_BLAS
 #define ARMA_DONT_USE_LAPACK
-
 #include <RcppArmadillo.h>
 
 #undef dnorm
 
 #include <cmath>
 #include <string>
+#include <functional>
 
 #include <cppbugs/cppbugs.hpp>
 
 #include <boost/random/exponential_distribution.hpp>
 #include <cstdlib>
 
+#ifdef SUPPORT_OPENMP
+#include<omp.h>
+#endif
+
 #include <progress.hpp>
 
 using namespace std;
 using namespace arma;
 using namespace cppbugs;
 
-/* A fast, yet precise, exp approximation */
-const int exp_precision = 10;
-double expTable[1<<exp_precision];
-void expFastInit(void) {
-  for(int i = 0; i < (1<<exp_precision); i++)
-    expTable[i] = pow(2, 1.*i/(1<<exp_precision));
-}
-
-inline double exp_approx(double x) {
-  x *= 1.44269504088896340737*(1<<exp_precision);
-  if(x >= 1024*(1<<exp_precision))
-    return INFINITY;
-  if(x < -1074*(1<<exp_precision))
-    return 0;
-  int xQuant = lrint(x);
-  double xQuantD = xQuant;
-  long long exp = (xQuant >> exp_precision)+1023;
-  union { double dbl; long long ll; } conv;
-  conv.ll = exp <= 0 ? 1LL<<(exp+51) : exp<<52;
-  return conv.dbl * expTable[xQuant & ((1<<exp_precision)-1)] * ((x-xQuantD)*(1/(1.44269504088896340737*(1<<exp_precision)))+1);
-}
-
-namespace arma {
-  // exp_approx
-  class eop_exp_approx: public eop_core<eop_exp_approx> {};
-
-  template<> template<typename eT> arma_hot arma_pure arma_inline eT
-  eop_core<eop_exp_approx>::process(const eT val, const eT  ) {
-    return exp_approx(val);
-  }
-
-  // Base
-  template<typename T1>
-  arma_inline
-  const eOp<T1, eop_exp_approx> exp_approx(const Base<typename T1::elem_type,T1>& A) {
-    arma_extra_debug_sigprint();
-    return eOp<T1, eop_exp_approx>(A.get_ref());
-  }
-
-  // BaseCube
-  template<typename T1>
-  arma_inline
-  const eOpCube<T1, eop_exp_approx> exp_approx(const BaseCube<typename T1::elem_type,T1>& A) {
-    arma_extra_debug_sigprint();
-    return eOpCube<T1, eop_exp_approx>(A.get_ref());
-  }
-}
-
 namespace dfmta {
 
 double TARG_SUP, EFF_MIN;
 bool HAS_TIME;
 double TIMEFULL, CYCLE;
 int COHORT_START, COHORT;
-std::minstd_rand r;
 
 /* True toxicity and response rates from input */
 struct true_data {
   vector<double> piV;
-  vector<vector<double> > respV;
+  vector<vector<double>> respV;
   vector<double> incl_per_week;
 };
 
@@ -93,10 +48,8 @@ struct toxicity_parameters {
   toxicity_parameters(): beta0(0), beta1(0) {}
 
   template<typename U>
-  U proba_tox(const U& dose_tox) {
-    U e = exp_approx(beta0 + beta1*dose_tox);
-    U p =  e/(1+e);
-    return p*(1-2e-10)+1e-10;
+  void proba_tox(const U& dose_tox, U& out) {
+    out = 1-1/(1+exp_approx(beta0 + beta1*dose_tox));
   }
 };
 
@@ -104,27 +57,16 @@ struct toxicity_parameters {
 struct efficacy_parameters {
   double gamma0, gamma1;
   int tau;
-  vec doseE;
 
-  efficacy_parameters(double gamma0, double gamma1, int tau, const vector<double>& doseE):
-    gamma0(gamma0), gamma1(gamma1), tau(tau), doseE(vec(doseE.data(), doseE.size())) {}
-  efficacy_parameters(const vector<double>& doseE):
-    gamma0(0), gamma1(0), tau(0), doseE(vec(doseE.data(), doseE.size())) {}
+  efficacy_parameters(double gamma0, double gamma1, int tau):
+    gamma0(gamma0), gamma1(gamma1), tau(tau) {}
+  efficacy_parameters():
+    gamma0(0), gamma1(0), tau(0) {}
 
-  vec responseRate(const uvec& id_dose) const {
-    unsigned tau = min((int)doseE.size()-1, max(0, this->tau));
-    uvec sup_tau = id_dose > tau;
-    vec e = exp_approx(gamma0 + gamma1*doseE.elem(sup_tau*tau+(1-sup_tau)%id_dose));
-    vec p =  e/(1+e);
-    return p*(1-2e-10)+1e-10;
-  }
-
-  double responseRate(unsigned id_dose) const {
-    unsigned tau = min((int)doseE.size()-1, max(0, this->tau));
-    unsigned sup_tau = id_dose > tau;
-    double e = exp_approx(gamma0 + gamma1*doseE(sup_tau*tau+(1-sup_tau)*id_dose));
-    double p =  e/(1+e);
-    return p*(1-2e-10)+1e-10;
+  template<typename U>
+  void responseRate(const vector<U>& dose_eff_tau, U& out) const {
+    unsigned tau = min((int)dose_eff_tau.size()-1, max(0, this->tau));
+    out = 1-1/(1+exp_approx(gamma0 + gamma1*dose_eff_tau[tau]));
   }
 };
 
@@ -138,17 +80,16 @@ struct estimations {
   vector<double> resp, qeff_inf, resp2;
   vector<double> proba_tau;
 
-  estimations(const vector<double>& doseE):
-    eff_params(doseE),
-    pi(doseE.size()), ptox_inf(doseE.size()), resp(doseE.size()),
-    qeff_inf(doseE.size()), resp2(doseE.size()), proba_tau(doseE.size())
+  estimations(int ndose):
+    pi(ndose), ptox_inf(ndose), resp(ndose),
+    qeff_inf(ndose), resp2(ndose), proba_tau(ndose)
   { }
 };
 
 
 struct trial_data;
 
-  typedef estimations (*estimator)(const trial_data& trial_data, int group, bool final);
+typedef estimations (*estimator)(const trial_data& trial_data, uint_fast64_t seed, int group, bool final);
 
 /* All the data needed dring a trial */
 struct trial_data{
@@ -156,7 +97,7 @@ struct trial_data{
 
   /* Transformation of dose levels to a more appropriate scale for estimation */
   vector<double> doseT;
-  vector<vector<double> > doseE;
+  vector<vector<double>> doseE;
 
   /* State information for inclusion process */
   vector<int> cdose; /* Current dose. -1 if stopped because of too high toxivity */
@@ -164,7 +105,7 @@ struct trial_data{
   double time_cur;
   int pat_incl;
   vector<int> pat_incl_group;
-  vector<unsigned> id_dose;
+  vector<unsigned> dose_adm;
   vector<int> group;
   vector<double> time_eff, time_incl;
   vector<int> efficacy, toxicity;
@@ -173,16 +114,18 @@ struct trial_data{
 
   vector<double> s_1;
 
-  trial_data(estimator est, const vector<double> doseTV, const vector<vector<double> > doseEV,
-             int ngroups, double s_2,
-             vector<double> s_1):
+  std::mt19937_64 r;
+
+  trial_data(estimator est, const vector<double>& doseTV, const vector<vector<double>>& doseEV,
+             int ngroups, double s_2, vector<double> s_1, uint_fast64_t seed):
     est(est),
     doseT(doseTV.size()), doseE(doseEV.size(), vector<double>(ngroups)),
     cdose(ngroups, 0), startup_end(ngroups, -1),
     time_cur(0), pat_incl(0),
     pat_incl_group(ngroups, 0),
     s_2(s_2),
-    s_1(s_1)
+    s_1(s_1),
+    r(seed)
   {
     for(int i=0; i < (int)doseT.size(); i++) {
       doseT[i] = log(doseTV[i]/(1-doseTV[i]));
@@ -195,9 +138,9 @@ struct trial_data{
 /* Results of a set of trials */
 struct results {
   vector<int> inconc;
-  vector<vector<int> > n_pat_dose, rec_dose;
+  vector<vector<int>> n_pat_dose, rec_dose;
   int n_pat_tot;
-  vector<vector<int> > n_tox, n_eff;
+  vector<vector<int>> n_tox, n_eff;
   int tox_tot, eff_tot;
   vector<int> n_pat_mtd;
   double duration;
@@ -210,12 +153,18 @@ struct results {
     n_tox(ndose, vector<int>(ngroups, 0)),      n_eff(ndose, vector<int>(ngroups, 0)),
     tox_tot(0), eff_tot(0),
     n_pat_mtd(ngroups, 0),
-    duration(0), nb_trials(0) {
-  }
+    duration(0), nb_trials(0)
+  { }
 
-  void accumulate(const trial_data& trial_data) {
-    for(int pat=0; pat<trial_data.pat_incl; pat++){
-      int dose_pat = trial_data.id_dose[pat];
+  void accumulate(const trial_data& trial_data, const vector<int>& recom) {
+    for(int group=0; group<recom.size(); group++)
+      if(recom[group] == -1)
+        inconc[group]++;
+      else
+        rec_dose[recom[group]][group]++;
+
+    for(int pat=0; pat<trial_data.pat_incl; pat++) {
+      int dose_pat = trial_data.dose_adm[pat];
       int group_pat = trial_data.group[pat];
       int tox = trial_data.toxicity[pat];
       int eff = HAS_TIME ? trial_data.time_eff[pat] < TIMEFULL
@@ -226,6 +175,8 @@ struct results {
       tox_tot += tox;
       n_eff[dose_pat][group_pat] += eff;
       eff_tot += eff;
+      if(dose_pat == recom[group_pat])
+        n_pat_mtd[group_pat]++;
     }
     if(HAS_TIME)
       duration += trial_data.time_cur;
@@ -244,7 +195,7 @@ bool wait_patient(trial_data& trial_data, const true_data& true_data, int& group
   const int ngroups = trial_data.cdose.size();
   vector<double> hist(ngroups);
   for(int g = 0; g < ngroups; g++)
-    if(trial_data.cdose[g] == -1)
+    if(trial_data.cdose[g] < 0)
       // We do not chose this group, either because it is too toxic,
       // or because we have included all the patients of this group.
       hist[g] = 0;
@@ -265,18 +216,18 @@ bool wait_patient(trial_data& trial_data, const true_data& true_data, int& group
     std::discrete_distribution<int> group_rng(hist.begin(), hist.end());
 
     if(!HAS_TIME)
-      group = group_rng(r);
+      group = group_rng(trial_data.r);
     else {
       reject_patient:
 
       /* How much time do we wait for this patient ? */
-      double time_incl = exp_rng(r, rate);
+      double time_incl = exp_rng(trial_data.r, rate);
 
       /* We shift everything by this amount of time. */
       trial_data.time_cur += time_incl;
 
       /* What is the group of the new patient ? */
-      group = group_rng(r);
+      group = group_rng(trial_data.r);
 
       /* Should we reject this patient ? */
       int startup_end = trial_data.startup_end[group];
@@ -296,52 +247,45 @@ bool wait_patient(trial_data& trial_data, const true_data& true_data, int& group
 }
 
 void include_patient(trial_data& trial_data, const true_data& true_data, int group) {
-
-  trial_data.id_dose.push_back(trial_data.cdose[group]);
+  trial_data.dose_adm.push_back(trial_data.cdose[group]);
   trial_data.group.push_back(group);
 
   /* Is there a toxicity ? */
   double pi_ij = true_data.piV[trial_data.cdose[group]];
   std::uniform_real_distribution<double> uni_rng;
-  trial_data.toxicity.push_back(uni_rng(r) < pi_ij);
+  trial_data.toxicity.push_back(uni_rng(trial_data.r) < pi_ij);
 
   /* When will there be a efficacy ? */
   double resp_ij = true_data.respV[trial_data.cdose[group]][group];
   if(HAS_TIME) {
     boost::exponential_distribution<double> exp_rng;
     double param_exp = -log(1-resp_ij)/TIMEFULL;
-    trial_data.time_eff.push_back(exp_rng(r, param_exp));
+    trial_data.time_eff.push_back(exp_rng(trial_data.r, param_exp));
 
     trial_data.time_incl.push_back(trial_data.time_cur);
   } else
-    trial_data.efficacy.push_back(uni_rng(r) < resp_ij);
+    trial_data.efficacy.push_back(uni_rng(trial_data.r) < resp_ij);
 
   trial_data.pat_incl++;
   trial_data.pat_incl_group[group]++;
 }
 
-double mean(vector<double> v) {
-  return std::accumulate(v.begin(), v.end(), 0.0) / v.size();
-}
-
-
 template<typename T>
 typename T::value_type median(const T& x) {
   if(x.size() == 0)
     return 0;
   std::vector<typename T::value_type> v(x.begin(), x.end());
-  std::sort (v.begin(), v.end());
-  if(v.size() % 2 == 1){
+  std::nth_element(v.begin(), v.begin()+v.size()/2, v.end());
+  if(v.size() % 2 == 1)
     return v[v.size()/2];
-  }
-  else{
-    return (v[v.size()/2] + v[v.size()/2-1])/2.0;
-  }
+  else
+    return (v[v.size()/2] +
+            *std::max_element(v.begin(), v.begin()+v.size()/2))/2.;
 }
 
 /* Estimate toxicity and efficacy parameters.
    The efficacy is estimated only for the given group. */
-estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
+estimations estimate_ra(const trial_data& trial_data, uint_fast64_t seed, int group, bool final) {
   const int ndose = trial_data.doseT.size();
 
   vector<double> doseE(ndose);
@@ -361,13 +305,13 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
     conv_to<vec>::from(ivec(trial_data.toxicity.data(), NPatientsCycle));
   const vec dose_tox =
     vec(trial_data.doseT.data(), ndose)
-    .elem(uvec(trial_data.id_dose.data(), NPatientsCycle));
+    .elem(uvec(trial_data.dose_adm.data(), NPatientsCycle));
 
   uvec in_group = find(ivec(trial_data.group.data(), NPatientsCycle) == group);
-  vec efficacy, time_min_eff, weights;
+  vec efficacy, weights;
   if(HAS_TIME) {
     efficacy = vec(in_group.size());
-    time_min_eff = vec(in_group.size());
+    vec time_min_eff(in_group.size());
 
     int card2 = 0;
     for(unsigned i = 0; i < in_group.size(); i++) {
@@ -401,28 +345,34 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
   } else
     efficacy = conv_to<vec>::from(ivec(trial_data.efficacy.data(), NPatientsCycle).elem(in_group));
 
-  const uvec id_dose_eff = uvec(trial_data.id_dose.data(), NPatientsCycle).elem(in_group);
+  const uvec dose_adm_eff = uvec(trial_data.dose_adm.data(), NPatientsCycle).elem(in_group);
+
+  vector<vec> dose_eff_tau;
+  for(int tau = 0; tau < ndose; tau++)
+    dose_eff_tau.push_back(vec(doseE.data(), ndose)
+                           .elem(clamp(dose_adm_eff, 0, tau)));
 
   /* MCMC initalization and sampling */
-  efficacy_parameters eff_params(0, 1, ndose-1, doseE);
-  toxicity_parameters tox_params(0, 1);
+  efficacy_parameters eff_params(0, 0.001, ndose-1);
+  toxicity_parameters tox_params(0, 0.001);
 
   vec pi_jk, resp_jk;
 
   std::function<void ()> model = [&]() {
-    if(HAS_TIME)
-      resp_jk = weights%eff_params.responseRate(id_dose_eff);
-    else
-      resp_jk = eff_params.responseRate(id_dose_eff);
-    pi_jk = tox_params.proba_tox(dose_tox);
+    if(HAS_TIME) {
+      eff_params.responseRate(dose_eff_tau, resp_jk);
+      resp_jk = weights % resp_jk;
+    } else
+      eff_params.responseRate(dose_eff_tau, resp_jk);
+    tox_params.proba_tox(dose_tox, pi_jk);
   };
-  MCModel<std::minstd_rand> m(model);
+  MCModel<std::mt19937> m(model);
 
   m.track<Normal>(eff_params.gamma0).dnorm(0, 0.01);
   m.track<Exponential>(eff_params.gamma1).dexp(1);
   vec distr(ndose);
   for(int i = 0; i < ndose; i++)
-    distr(i) = pow(1.0,i);
+    distr[i] = 1.0;
   m.track<Discrete>(eff_params.tau).ddiscr(distr);
   m.track<ObservedBernoulli>(efficacy).dbern(resp_jk);
 
@@ -433,7 +383,7 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
   m.sample(1e6, 1e5, 1e4, 10);
 
   /* Get list of eff_params and tox_params */
-  vector<pair<efficacy_parameters, toxicity_parameters> > params_draws_list;
+  vector<pair<efficacy_parameters, toxicity_parameters>> params_draws_list;
   auto itbeta0 = m.getNode(tox_params.beta0).history.begin(),
     itbeta0end = m.getNode(tox_params.beta0).history.end();
   auto itbeta1 = m.getNode(tox_params.beta1).history.begin();
@@ -442,9 +392,9 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
   auto ittau = m.getNode(eff_params.tau).history.begin();
 
   while(itbeta0 != itbeta0end) {
-    params_draws_list.push_back(make_pair(
-       efficacy_parameters(*itgamma0, *itgamma1, *ittau, doseE),
-       toxicity_parameters(*itbeta0, *itbeta1)));
+    params_draws_list.emplace_back(
+      efficacy_parameters(*itgamma0, *itgamma1, *ittau),
+      toxicity_parameters(*itbeta0, *itbeta1));
     itbeta0++; itbeta1++;
     itgamma0++; itgamma1++; ittau++;
   }
@@ -452,7 +402,7 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
   int n_samp = params_draws_list.size();
 
   /* Estimations */
-  estimations result(doseE);
+  estimations result(ndose);
   result.tox_params.beta0 = median(m.getNode(tox_params.beta0).history);
   result.tox_params.beta1 = median(m.getNode(tox_params.beta1).history);
 
@@ -460,7 +410,8 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
     int count_tox = 0;
     std::vector <double> pi_median;
     for(auto draw: params_draws_list){
-      double proba_tox = draw.second.proba_tox(trial_data.doseT[d]);
+      double proba_tox;
+      draw.second.proba_tox(trial_data.doseT[d], proba_tox);
       count_tox += proba_tox < TARG_SUP;
       pi_median.push_back(proba_tox);
     }
@@ -489,6 +440,7 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
       }
     }
     std::discrete_distribution<int> tau_rng(prob_tau_high.begin(), prob_tau_high.end());
+    mt19937_64 r(seed);
     int sample = tau_rng(r);
     result.eff_params.tau = tau_high_prob[sample];
   }
@@ -499,8 +451,12 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
   for(int d = 0; d < ndose; d++) {
     int count_eff = 0;
     std::vector <double> q_median;
+    vector<double> dose_eff_tau;
+    for(int tau = 0; tau < ndose; tau++)
+      dose_eff_tau.push_back(doseE[min(tau, d)]);
     for(auto draw: params_draws_list){
-      double resp_rate = draw.first.responseRate(d);
+      double resp_rate;
+      draw.first.responseRate(dose_eff_tau, resp_rate);
       count_eff += resp_rate < EFF_MIN;
       q_median.push_back(resp_rate);
     }
@@ -519,7 +475,7 @@ estimations estimate_ra(const trial_data& trial_data, int group, bool final) {
 
 /* Estimate toxicity and efficacy parameters.
    The efficacy is estimated only for the given group. */
-estimations estimate_pm(const trial_data& trial_data, int group, bool) {
+estimations estimate_pm(const trial_data& trial_data, uint_fast64_t, int group, bool) {
   const int ndose = trial_data.doseT.size();
 
   if(!HAS_TIME)
@@ -539,7 +495,7 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
     conv_to<vec>::from(ivec(trial_data.toxicity.data(), NPatientsCycle));
   const vec dose_tox =
     vec(trial_data.doseT.data(), ndose)
-    .elem(uvec(trial_data.id_dose.data(), NPatientsCycle));
+    .elem(uvec(trial_data.dose_adm.data(), NPatientsCycle));
 
   uvec in_group = find(ivec(trial_data.group.data(), NPatientsCycle) == group);
   vec efficacy(in_group.size()), time_min_eff(in_group.size());
@@ -573,25 +529,31 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
       weights(i) = (card1+w_ini)/(card2+1);
     }
   }
-  const uvec id_dose_eff = uvec(trial_data.id_dose.data(), NPatientsCycle).elem(in_group);
+  const uvec dose_adm_eff = uvec(trial_data.dose_adm.data(), NPatientsCycle).elem(in_group);
+
+  vector<vec> dose_eff_tau;
+  for(int tau = 0; tau < ndose; tau++)
+    dose_eff_tau.push_back(vec(doseE.data(), ndose)
+                           .elem(clamp(dose_adm_eff, 0, tau)));
 
   /* MCMC initalization and sampling */
-  efficacy_parameters eff_params(0, 1, ndose-1, doseE);
-  toxicity_parameters tox_params(0, 1);
+  efficacy_parameters eff_params(0, 0.001, ndose-1);
+  toxicity_parameters tox_params(0, 0.001);
 
-  vec pi_jk, resp_jk_weight;
+  vec pi_jk, resp_jk, resp_jk_weight;
 
   std::function<void ()> model = [&]() {
-    resp_jk_weight = weights%eff_params.responseRate(id_dose_eff);
-    pi_jk = tox_params.proba_tox(dose_tox);
+    eff_params.responseRate(dose_eff_tau, resp_jk);
+    resp_jk_weight = weights % resp_jk;
+    tox_params.proba_tox(dose_tox, pi_jk);
   };
-  MCModel<std::minstd_rand> m(model);
+  MCModel<std::mt19937> m(model);
 
   m.track<Normal>(eff_params.gamma0).dnorm(0, 0.01);
   m.track<Exponential>(eff_params.gamma1).dexp(1);
   vec distr(ndose);
   for(int i = 0; i < ndose; i++)
-    distr(i) = pow(1.0,i);
+    distr[i] = 1.0;
   m.track<Discrete>(eff_params.tau).ddiscr(distr);
   m.track<ObservedBernoulli>(efficacy).dbern(resp_jk_weight);
 
@@ -602,7 +564,7 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
   m.sample(1e6, 1e5, 1e4, 10);
 
   /* Get list of eff_params and tox_params */
-  vector<pair<efficacy_parameters, toxicity_parameters> > params_draws_list;
+  vector<pair<efficacy_parameters, toxicity_parameters>> params_draws_list;
   auto itbeta0 = m.getNode(tox_params.beta0).history.begin(),
     itbeta0end = m.getNode(tox_params.beta0).history.end();
   auto itbeta1 = m.getNode(tox_params.beta1).history.begin();
@@ -611,9 +573,9 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
   auto ittau = m.getNode(eff_params.tau).history.begin();
 
   while(itbeta0 != itbeta0end) {
-    params_draws_list.push_back(make_pair(
-       efficacy_parameters(*itgamma0, *itgamma1, *ittau, doseE),
-       toxicity_parameters(*itbeta0, *itbeta1)));
+    params_draws_list.emplace_back(
+      efficacy_parameters(*itgamma0, *itgamma1, *ittau),
+      toxicity_parameters(*itbeta0, *itbeta1));
     itbeta0++; itbeta1++;
     itgamma0++; itgamma1++; ittau++;
   }
@@ -622,7 +584,7 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
 
   /* Estimations */
   /* Estimation of toxicity */
-  estimations result(doseE);
+  estimations result(ndose);
   result.tox_params.beta0 = median(m.getNode(tox_params.beta0).history);
   result.tox_params.beta1 = median(m.getNode(tox_params.beta1).history);
 
@@ -630,7 +592,8 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
     int count_tox = 0;
     std::vector <double> pi_median;
     for(auto draw: params_draws_list){
-      double proba_tox = draw.second.proba_tox(trial_data.doseT[d]);
+      double proba_tox;
+      draw.second.proba_tox(trial_data.doseT[d], proba_tox);
       count_tox += proba_tox < TARG_SUP;
       pi_median.push_back(proba_tox);
     }
@@ -639,8 +602,8 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
   }
 
   /* Estimation of efficacy */
-  vector<vector<double> > qeff_inf_tau(ndose, vector<double>(ndose));
-  vector<vector<double> > resp_tau(ndose, vector<double>(ndose));
+  vector<vector<double>> qeff_inf_tau(ndose, vector<double>(ndose));
+  vector<vector<double>> resp_tau(ndose, vector<double>(ndose));
   for(int tau = 0; tau < ndose; tau++) {
     int count = 0;
     vector<double> gammas0, gammas1;
@@ -658,9 +621,13 @@ estimations estimate_pm(const trial_data& trial_data, int group, bool) {
       for(int d = 0; d < ndose; d++) {
         int count_eff = 0;
         std::vector <double> q_median;
+        vector<double> dose_eff_tau;
+        for(int tau = 0; tau < ndose; tau++)
+          dose_eff_tau.push_back(doseE[min(tau, d)]);
         for(auto draw: params_draws_list){
           if(tau == draw.first.tau) {
-            double resp_rate = draw.first.responseRate(d);
+            double resp_rate;
+            draw.first.responseRate(dose_eff_tau, resp_rate);
             count_eff += resp_rate < EFF_MIN;
             q_median.push_back(resp_rate);
           }
@@ -725,6 +692,11 @@ int find_next_dose(trial_data& trial_data, int group, double c_tox, double c_eff
                    bool final, estimations* estimRet = NULL){
   const int ndose = trial_data.doseT.size();
 
+  if(trial_data.startup_end[group] == -1 && final)
+    trial_data.startup_end[group] = trial_data.pat_incl;
+
+  // Warning : this code should detect whether the startup has ended, even
+  // if it was a long time ago.
   if(trial_data.startup_end[group] == -1) {
     if(trial_data.pat_incl_group[group] == 0)
       return 0;
@@ -732,7 +704,7 @@ int find_next_dose(trial_data& trial_data, int group, double c_tox, double c_eff
     int dlt = 0;
     for(int pat = 0; pat < trial_data.pat_incl; pat++)
       if(trial_data.group[pat] == group &&
-         trial_data.id_dose[pat] == trial_data.cdose[group] &&
+         trial_data.dose_adm[pat] == trial_data.cdose[group] &&
          trial_data.toxicity[pat])
         dlt++;
 
@@ -742,7 +714,7 @@ int find_next_dose(trial_data& trial_data, int group, double c_tox, double c_eff
     trial_data.startup_end[group] = trial_data.pat_incl_group[group];
   }
 
-  estimations estim = trial_data.est(trial_data, group, final);
+  estimations estim = trial_data.est(trial_data, trial_data.r(), group, final);
   if(estimRet != NULL)
     *estimRet = estim;
 
@@ -753,9 +725,9 @@ int find_next_dose(trial_data& trial_data, int group, double c_tox, double c_eff
   int highest_tested=0;
   for(int pat = 0; pat < trial_data.pat_incl; pat++){
     if(trial_data.group[pat] == group){
-      pat_explo_dose[trial_data.id_dose[pat]]++;
-      if((int)trial_data.id_dose[pat] > highest_tested){
-        highest_tested = trial_data.id_dose[pat];
+      pat_explo_dose[trial_data.dose_adm[pat]]++;
+      if((int)trial_data.dose_adm[pat] > highest_tested){
+        highest_tested = trial_data.dose_adm[pat];
       }
     }
   }
@@ -773,10 +745,8 @@ int find_next_dose(trial_data& trial_data, int group, double c_tox, double c_eff
   }
 
   int max_candidate_dose;
-  if(final)
-    max_candidate_dose = highest_tested;
-  else
-    max_candidate_dose = max(min(ndose-1, cdose+1), highest_tested);
+  if(final) max_candidate_dose = highest_tested;
+  else max_candidate_dose = max(min(ndose-1, cdose+1), highest_tested);
 
   for(int candidate_dose = 0; candidate_dose <= max_candidate_dose; candidate_dose++) {
     // Determination of the next dose or stopping rule
@@ -811,7 +781,6 @@ static R_NativePrimitiveArgType dfmta_next_args[] =
    REALSXP, REALSXP,
    INTSXP,
 
-   INTSXP,
    INTSXP,
    REALSXP,
    INTSXP,
@@ -820,6 +789,7 @@ static R_NativePrimitiveArgType dfmta_next_args[] =
    LGLSXP, LGLSXP,
    LGLSXP,
 
+   LGLSXP,
    REALSXP, REALSXP,
    REALSXP, REALSXP,
    REALSXP };
@@ -836,14 +806,14 @@ static void dfmta_next(int* has_time, int* ra_pm /* true for RA, false for PM */
                        int* seed,
 
                        int* cdose /* -1 if group stopped */,
-                       int* startup_end /* -1 if startup has not ended */,
                        double* time_cur,
                        int* pat_incl_group,
-                       int* id_dose, int* group,
+                       int* dose_adm, int* group,
                        double* time_eff, double* time_incl,
                        int* efficacy, int* toxicity,
                        int* final,
 
+                       int* in_startup,
                        double* pi, double* ptox_inf,
                        double* resp2, double* qeff_inf,
                        double* proba_tau) {
@@ -861,30 +831,29 @@ static void dfmta_next(int* has_time, int* ra_pm /* true for RA, false for PM */
   int group_cur = group[pat_incl];
 
   vector<double> doseTV(doseTV0, doseTV0+*ndose);
-  vector<vector<double> > doseEV(*ndose, vector<double>(*ngroups, 0));
+  vector<vector<double>> doseEV(*ndose, vector<double>(*ngroups, 0));
   for(int i = 0; i < *ndose; i++)
     doseEV[i][group_cur] = doseEV0[i];
 
   COHORT_START = *cohort_start;
   COHORT = *cohort;
 
-  r.seed(*seed);
-
   estimator est = *ra_pm ? estimate_ra : estimate_pm;
   vector<double> _s_1;
   if(!*final) {
     _s_1.resize(pat_incl_group[group_cur]+1);
     _s_1[pat_incl_group[group_cur]] = *s_1;
   }
-  trial_data trial_data(est, doseTV, doseEV, *ngroups, *s_2, _s_1);
+  trial_data trial_data(est, doseTV, doseEV, *ngroups, *s_2, _s_1, *seed);
   trial_data.pat_incl = pat_incl;
   trial_data.cdose = vector<int>(*ngroups, 0);
   trial_data.cdose[group_cur] = *cdose;
   trial_data.startup_end = vector<int>(*ngroups, 0);
-  trial_data.startup_end[group_cur] = *startup_end;
+  // Assume the startup has not ended : if it has, it will detect it.
+  trial_data.startup_end[group_cur] = -1;
   trial_data.time_cur = *time_cur;
   trial_data.pat_incl_group = vector<int>(pat_incl_group, pat_incl_group+*ngroups);
-  trial_data.id_dose = vector<unsigned int>(id_dose, id_dose+pat_incl);
+  trial_data.dose_adm = vector<unsigned int>(dose_adm, dose_adm+pat_incl);
   trial_data.group = vector<int>(group, group+pat_incl);
   if(HAS_TIME) {
     trial_data.time_eff = vector<double>(time_eff, time_eff+pat_incl);
@@ -894,12 +863,11 @@ static void dfmta_next(int* has_time, int* ra_pm /* true for RA, false for PM */
   }
   trial_data.toxicity = vector<int>(toxicity, toxicity+pat_incl);
 
-  vector<double> dummy;
-  estimations estim(dummy);
+  estimations estim(*ndose);
   *cdose = find_next_dose(trial_data, group_cur, *c_tox, *c_eff, *final, &estim);
-  *startup_end = trial_data.startup_end[group_cur];
+  *in_startup = trial_data.startup_end[group_cur] == -1;
 
-  if(startup_end[group_cur] != -1)
+  if(!in_startup)
     for(int d = 0; d < *ndose; d++) {
       pi[d] = estim.pi[d];
       ptox_inf[d] = estim.ptox_inf[d];
@@ -927,13 +895,13 @@ static R_NativePrimitiveArgType dfmta_simu_args[] =
    INTSXP, INTSXP, INTSXP,
    REALSXP, REALSXP,
    REALSXP, REALSXP,
-   INTSXP,
+   INTSXP, INTSXP,
 
    INTSXP, INTSXP, INTSXP,
    INTSXP, INTSXP, INTSXP,
    INTSXP, INTSXP, INTSXP,
    REALSXP};
-static const int dfmta_simu_nargs = 32;
+static const int dfmta_simu_nargs = 33;
 
 static void dfmta_simu(int* has_time, int* ra_pm /* true for RA, false for PM */,
                        int* ngroups, int* ndose,
@@ -945,123 +913,139 @@ static void dfmta_simu(int* has_time, int* ra_pm /* true for RA, false for PM */
                        int* cohort_start, int* cohort, int* ntrial,
                        double* c_tox, double* c_eff,
                        double* s_1, double* s_2,
-                       int* seed,
+                       int* seed, int* nthreads,
 
                        int* inconc, int* n_pat_dose, int* rec_dose,
                        int* n_pat_tot, int* n_tox, int* n_eff,
                        int* tox_tot, int* eff_tot, int* n_pat_mtd,
                        double* duration) {
-  try {
+  string errstr;
 
-  struct true_data true_data;
+  {
+    struct true_data true_data;
 
-  CYCLE = *cycle;
-  HAS_TIME = *has_time;
-  if(HAS_TIME) TIMEFULL = *time_full;
+    CYCLE = *cycle;
+    HAS_TIME = *has_time;
+    if(HAS_TIME) TIMEFULL = *time_full;
 
-  vector<double> _s_1(s_1, s_1+*npatientspergroup);
-  true_data.piV = vector<double>(piV, piV+*ndose);
+    vector<double> _s_1(s_1, s_1+*npatientspergroup);
+    true_data.piV = vector<double>(piV, piV+*ndose);
 
-  true_data.respV = vector<vector<double> >(*ndose, vector<double>(*ngroups));
-  for(int g = 0; g < *ngroups; g++)
-    for(int i = 0; i < *ndose; i++)
-      true_data.respV[i][g] = respV[i**ngroups + g];
+    true_data.respV = vector<vector<double> >(*ndose, vector<double>(*ngroups));
+    for(int g = 0; g < *ngroups; g++)
+      for(int i = 0; i < *ndose; i++)
+        true_data.respV[i][g] = respV[i**ngroups + g];
 
-  TARG_SUP = *targ_sup;
-  EFF_MIN = *eff_min;
+    TARG_SUP = *targ_sup;
+    EFF_MIN = *eff_min;
 
-  vector<double> doseTV(doseTV0, doseTV0+*ndose);
-  vector<vector<double> > doseEV(*ndose, vector<double>(*ngroups, 0));
-  for(int g = 0; g < *ngroups; g++)
-    for(int i = 0; i < *ndose; i++)
-      doseEV[i][g] = doseEV0[i**ngroups + g];
-  true_data.incl_per_week = vector<double>(incl_per_week, incl_per_week+*ngroups);
+    vector<double> doseTV(doseTV0, doseTV0+*ndose);
+    vector<vector<double> > doseEV(*ndose, vector<double>(*ngroups, 0));
+    for(int g = 0; g < *ngroups; g++)
+      for(int i = 0; i < *ndose; i++)
+        doseEV[i][g] = doseEV0[i**ngroups + g];
+    true_data.incl_per_week = vector<double>(incl_per_week, incl_per_week+*ngroups);
 
-  COHORT_START = *cohort_start;
-  COHORT = *cohort;
+    COHORT_START = *cohort_start;
+    COHORT = *cohort;
 
-  r.seed(*seed);
+    std::mt19937_64 global_rng(*seed);
+    vector<uint_fast64_t> seeds;
+    for(int trial=0; trial<*ntrial; trial++)
+      seeds.push_back(global_rng());
 
-  struct results results(*ndose, *ngroups);
+    struct results results(*ndose, *ngroups);
 
-  Progress prog(*ntrial);
-  for(int trial=0; trial<*ntrial; trial++){
-    estimator est = *ra_pm ? estimate_ra : estimate_pm;
-    struct trial_data trial_data(est, doseTV, doseEV, *ngroups, *s_2, _s_1);
+#ifdef SUPPORT_OPENMP
+    if(*nthreads > 0)
+      omp_set_num_threads(*nthreads);
+    else
+      omp_set_num_threads(omp_get_num_procs());
+#endif
 
-    while(true) {
-      if(prog.check_abort()) {
-        *ntrial = trial;
-        goto aborted;
-      }
+    bool err = false;
+    Progress prog(*ntrial);
+#pragma omp parallel for schedule(dynamic, 1)
+    for(int trial=0; trial<*ntrial; trial++) {
+      try {
+        bool err2;
+#pragma omp critical
+        err2 = err;
+        if(prog.check_abort() || err2)
+          continue;
 
-      int group;
-      if(wait_patient(trial_data, true_data, group))
-        break;
+        estimator est = *ra_pm ? estimate_ra : estimate_pm;
+        struct trial_data trial_data(est, doseTV, doseEV, *ngroups, *s_2, _s_1, seeds[trial]);
 
-      int startup_end = trial_data.startup_end[group];
-      if( (startup_end == -1 && trial_data.pat_incl_group[group] % COHORT_START == 0) ||
-          (startup_end != -1 && (trial_data.pat_incl_group[group] - startup_end) % COHORT == 0) ) {
-        trial_data.cdose[group] = find_next_dose(trial_data, group, *c_tox, *c_eff, false);
-      }
+        while(true) {
+#pragma omp critical
+          err2 = err;
+          if(prog.check_abort() || err2)
+            goto aborted;
 
-      if(trial_data.cdose[group] != -1) {
-        include_patient(trial_data, true_data, group);
-        if(trial_data.pat_incl_group[group] > *npatientspergroup)
-          throw std::logic_error("Internal error: too many patients in a group.");
-        if(trial_data.pat_incl_group[group] == *npatientspergroup)
-          trial_data.cdose[group] = -1; /* No more patiens for this group */
-      }
-    }
+          int group;
+          if(wait_patient(trial_data, true_data, group))
+            break;
 
-    for(int group = 0; group < *ngroups; group++){
-      int recom = find_next_dose(trial_data, group, *c_tox, *c_eff, true);
-      if(recom == -1) {
-        results.inconc[group]++;
-      }
-      else {
-        results.rec_dose[recom][group]++;
-        for(int pat = 0; pat < trial_data.pat_incl; pat++){
-          if(trial_data.group[pat] == group &&
-             trial_data.id_dose[pat] == recom) {
-            results.n_pat_mtd[group]++;
+          int startup_end = trial_data.startup_end[group];
+          if( (startup_end == -1 && trial_data.pat_incl_group[group] % COHORT_START == 0) ||
+              (startup_end != -1 && (trial_data.pat_incl_group[group] - startup_end) % COHORT == 0) )
+            trial_data.cdose[group] = find_next_dose(trial_data, group, *c_tox, *c_eff, false);
+
+          if(trial_data.cdose[group] >= 0) {
+            include_patient(trial_data, true_data, group);
+            if(trial_data.pat_incl_group[group] >= *npatientspergroup)
+              trial_data.cdose[group] = -2; /* No more patiens for this group */
           }
         }
+
+        vector<int> recom;
+        for(int group = 0; group < *ngroups; group++)
+          if(trial_data.cdose[group] == -2)
+            recom.push_back(find_next_dose(trial_data, group, *c_tox, *c_eff, true));
+          else
+            recom.push_back(-1);
+
+#pragma omp critical
+        results.accumulate(trial_data, recom);
+
+      } catch (std::logic_error &e) {
+#pragma omp critical
+        { err = true;
+          errstr = e.what(); }
+      } catch(...) {
+#pragma omp critical
+        err = true;
       }
+
+#pragma omp critical
+      prog.increment();
+      aborted: ;
     }
-    results.accumulate(trial_data);
 
-    prog.increment();
-  }
+    if(err) goto errlbl;
 
-  aborted:
-
-  copy(results.inconc.begin(), results.inconc.end(), inconc);
-  for(int i = 0; i < *ndose; i++)
-    for(int j = 0; j < *ngroups; j++) {
-      n_pat_dose[i**ngroups + j] = results.n_pat_dose[i][j];
-      rec_dose[i**ngroups + j]   = results.rec_dose[i][j];
-      n_tox[i**ngroups + j]      = results.n_tox[i][j];
-      n_eff[i**ngroups + j]      = results.n_eff[i][j];
-    }
-  *n_pat_tot = results.n_pat_tot;
-  *tox_tot   = results.tox_tot;
-  *eff_tot   = results.eff_tot;
-  *duration  = results.duration;
-  copy(results.n_pat_mtd.begin(), results.n_pat_mtd.end(), n_pat_mtd);
-
-  if(results.nb_trials != *ntrial)
-    throw std::logic_error("Internal error: invalid ntrials.");
+    copy(results.inconc.begin(), results.inconc.end(), inconc);
+    for(int i = 0; i < *ndose; i++)
+      for(int j = 0; j < *ngroups; j++) {
+        n_pat_dose[i**ngroups + j] = results.n_pat_dose[i][j];
+        rec_dose[i**ngroups + j]   = results.rec_dose[i][j];
+        n_tox[i**ngroups + j]      = results.n_tox[i][j];
+        n_eff[i**ngroups + j]      = results.n_eff[i][j];
+      }
+    *n_pat_tot = results.n_pat_tot;
+    *tox_tot   = results.tox_tot;
+    *eff_tot   = results.eff_tot;
+    *duration  = results.duration;
+    *ntrial    = results.nb_trials;
+    copy(results.n_pat_mtd.begin(), results.n_pat_mtd.end(), n_pat_mtd);
 
   }
-  catch (std::logic_error &e) {
-    error("Internal error in dfmta (details: %s)", e.what());
-  }
-  catch (...) {
-    error("Internal error in dfmta");
-  }
 
-  return;
+  if(false) {
+  errlbl:
+    error("Internal error in dfmta (details: %s)", errstr.c_str());
+  }
 }
 
 static const R_CMethodDef cMethods[] = {
@@ -1074,7 +1058,6 @@ static const R_CMethodDef cMethods[] = {
 
 void R_init_dfmta(DllInfo *info) {
   R_registerRoutines(info, cMethods, NULL, NULL, NULL);
-  expFastInit();
 }
 
 }
